from unittest.mock import Mock, patch

from click.testing import CliRunner
import pytest

from anyscale.client.openapi_client import (
    CreateOTPReturnApiModel,
    RequestOTPReturnApiModel,
)
from anyscale.client.openapi_client.models import (
    CreateotpreturnapimodelResponse,
    RequestotpreturnapimodelResponse,
)
from anyscale.commands.login_commands import anyscale_login, anyscale_logout


@pytest.mark.parametrize(
    "webbrowser_error", [False, True],
)
def test_anyscale_login(webbrowser_error: bool) -> None:
    runner = CliRunner()

    mock_api_client = Mock()
    mock_api_client.create_one_time_password_api_v2_users_create_otp_token_get = Mock(
        return_value=CreateotpreturnapimodelResponse(
            result=CreateOTPReturnApiModel(
                url="https://console.anyscale.com/otp/fake-otp-token",
                otp="fake-otp-token",
            )
        )
    )
    mock_api_client.check_one_time_password_api_v2_users_request_otp_token_otp_get = Mock(
        return_value=RequestotpreturnapimodelResponse(
            result=RequestOTPReturnApiModel(
                valid=True, server_session_id="fake-session-id",
            )
        )
    )
    mock_auth_controller_set = Mock()
    if webbrowser_error:
        mock_webbrowser_open = Mock(side_effect=Exception("some error"))
    else:
        mock_webbrowser_open = Mock()

    with patch.multiple(
        "anyscale.commands.login_commands",
        get_unauthenticated_openapi_client=Mock(return_value=mock_api_client),
        webbrowser=Mock(),
    ), patch.multiple(
        "anyscale.commands.login_commands.AuthController", set=mock_auth_controller_set,
    ), patch.multiple(
        "anyscale.commands.login_commands.webbrowser", open=mock_webbrowser_open,
    ):
        runner.invoke(anyscale_login)

        mock_api_client.create_one_time_password_api_v2_users_create_otp_token_get.assert_called_once()
        mock_api_client.check_one_time_password_api_v2_users_request_otp_token_otp_get.assert_called_once_with(
            "fake-otp-token"
        )
        mock_auth_controller_set.assert_called_once_with("fake-session-id")
        mock_webbrowser_open.assert_called_once_with(
            "https://console.anyscale.com/otp/fake-otp-token"
        )


def test_anyscale_logout() -> None:
    runner = CliRunner()

    mock_auth_controller_remove = Mock()
    with patch.multiple(
        "anyscale.commands.login_commands.AuthController",
        remove=mock_auth_controller_remove,
    ):
        runner.invoke(anyscale_logout)
        mock_auth_controller_remove.assert_called_once_with(ask=False)



"""
This file holds all of the CLI commands for the "anyscale machines" path.
"""
import click
from rich.console import Console
from rich.table import Table

from anyscale.controllers.machines_controller import MachinesController


@click.group(
    "machine", help="Commands to interact with machines in Anyscale.",
)
def machines_cli() -> None:
    pass


@machines_cli.command(name="list", help="List machines registered to Anyscale.")
@click.option("--cloud_id", type=str, required=True, help="Provide a cloud ID.")
def list_machines(cloud_id: str,) -> None:
    machines_controller = MachinesController()
    output = machines_controller.list_machines(cloud_id=cloud_id)

    table = Table()
    table.add_column("ID")
    table.add_column("Host Name")
    table.add_column("Shape")
    table.add_column("Connection State")
    table.add_column("Allocation State")
    table.add_column("Cluster ID")
    table.add_column("Node ID")
    for m in output.machines:
        table.add_row(
            m.machine_id,
            m.hostname,
            m.machine_shape,
            m.connection_state,
            m.allocation_state,
            m.cluster_id,
            m.node_id,
        )

    console = Console()
    console.print(table)



from enum import Enum
from typing import Optional, Union

import click
from pydantic import BaseModel
from typing_extensions import Literal


class EntityType(str, Enum):
    ID = "ID"
    NAME = "NAME"


class Entity(BaseModel):
    type: EntityType


class IdBasedEntity(Entity):
    type: Literal[EntityType.ID] = EntityType.ID
    id: str


class NameBasedEntity(Entity):
    type: Literal[EntityType.NAME] = EntityType.NAME
    name: str
    version: Optional[int] = None


def format_inputs_to_entity(
    name: Optional[str], entity_id: Optional[str]
) -> Union[IdBasedEntity, NameBasedEntity]:
    """
    Share method for CLI commands to accept either the name of the id of an entity.
    """
    if int(bool(name)) + int(bool(entity_id)) != 1:
        raise click.ClickException("Please provide exactly one of: name, id.")
    elif name:
        return NameBasedEntity(name=name)
    elif entity_id:
        return IdBasedEntity(id=entity_id)
    else:
        raise click.ClickException("Please provide exactly one of: name, id.")



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse

import click
from pydantic import BaseModel, Extra, Field, root_validator, validator
import yaml

from anyscale.cli_logger import LogsLogger
from anyscale.cluster_compute import (
    get_cluster_compute_from_name,
    get_default_cluster_compute,
    register_compute_template,
)
from anyscale.cluster_env import (
    get_build_from_cluster_env_identifier,
    get_default_cluster_env_build,
    validate_successful_build,
)
from anyscale.project import (
    get_parent_cloud_id_and_name_of_project,
    get_proj_id_from_name,
)
from anyscale.shared_anyscale_utils.utils.byod import BYODInfo
from anyscale.util import (
    extract_versions_from_image_name,
    PROJECT_NAME_ENV_VAR,
)


log = LogsLogger()


def _validate_conda_option(conda_option: Union[str, Dict]) -> Union[str, Dict]:
    """Parses and validates a user-provided 'conda' option.

    Can be one of three cases:
        1) A str that's the name of a pre-installed conda environment.
        2) A string pointing to a local conda environment YAML. In this case,
           the file contents will be read into a dict.
        3) A dict that defines a conda environment. This is passed through.
    """
    result = None
    if isinstance(conda_option, str):
        yaml_file = Path(conda_option)
        if yaml_file.suffix in (".yaml", ".yml"):
            if not yaml_file.is_file():
                raise click.ClickException(f"Can't find conda YAML file {yaml_file}.")
            try:
                result = yaml.safe_load(yaml_file.read_text())
            except Exception as e:  # noqa: BLE001
                raise click.ClickException(
                    f"Failed to read conda file {yaml_file}: {e}."
                )
        else:
            # Assume it's a pre-existing conda environment name.
            result = conda_option
    elif isinstance(conda_option, dict):
        result = conda_option

    return result


def _validate_pip_option(pip_option: Union[str, List[str]]) -> Optional[List[str]]:
    """Parses and validates a user-provided 'pip' option.

    Can be one of two cases:
        1) A List[str] describing the requirements. This is passed through.
        2) A string pointing to a local requirements file. In this case, the
           file contents will be read split into a list.
    """
    result = None
    if isinstance(pip_option, str):
        # We have been given a path to a requirements.txt file.
        pip_file = Path(pip_option)
        if not pip_file.is_file():
            raise click.ClickException(f"{pip_file} is not a valid file.")
        result = pip_file.read_text().strip().split("\n")
    elif isinstance(pip_option, list) and all(
        isinstance(dep, str) for dep in pip_option
    ):
        if len(pip_option) == 0:
            result = None
        else:
            result = pip_option

    return result


def _validate_py_modules(py_modules_option: List[str]) -> List[str]:
    for entry in py_modules_option:
        if "://" not in entry:
            raise click.ClickException(
                "Only remote URIs are currently supported for py_modules in the job "
                "config (not local directories). Please see "
                "https://docs.ray.io/en/master/handling-dependencies.html#remote-uris for supported options."
            )

    return py_modules_option


def _working_dir_is_remote_uri(working_dir: str) -> bool:
    return "://" in working_dir


def _validate_working_dir(working_dir_option: str) -> str:
    """If working_dir is a local directory, check that it exists."""
    # We have been given a path to a local directory.
    if (
        not _working_dir_is_remote_uri(working_dir_option)
        and not Path(working_dir_option).is_dir()
    ):
        raise click.ClickException(
            f"working_dir {working_dir_option} is not a valid local directory or remote URI."
        )

    return working_dir_option


def _validate_working_dir_and_upload_path(
    working_dir: Optional[str], upload_path: Optional[str]
):
    """Check that the combination of working_dir and upload_path is valid.

    Exception should be thrown if both working dir is a remote uri and upload path is defined
    Otherwise, all other permutations are valid
    """
    if (
        upload_path is not None
        and working_dir
        and _working_dir_is_remote_uri(working_dir)
    ):
        raise click.ClickException(
            f"`upload_path` was specified, but `working_dir` is not a local directory.  Recieved `upload_path`: {upload_path} and `working_dir`: {working_dir}."
        )
    elif upload_path is not None and not working_dir:
        raise click.ClickException(
            f"`upload_path` was specified, but no `working_dir` is defined. Recieved `upload_path`: {upload_path} and `working_dir`: None."
        )


def _validate_upload_path(upload_path: str) -> str:
    """Check that the upload path is a valid S3 or GS remote URI."""
    try:
        parsed_upload_path = urlparse(upload_path)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Failed to parse `upload_path` {upload_path} as a URI (e.g. 's3://my_bucket/my_dir'): {e}"
        )
    if parsed_upload_path.scheme not in ["s3", "gs"]:
        raise click.ClickException(
            f"Only Amazon S3 (e.g. 's3://bucket', 's3://bucket/path') and Google Storage URIs (e.g. 'gs://bucket', 'gs://bucket/path') are supported. Received {upload_path}."
        )
    return upload_path


def _validate_env_vars(env_vars: Dict[str, str]) -> Dict[str, str]:
    for key, value in env_vars.items():
        if not isinstance(key, str):
            raise click.ClickException(
                f"env_vars key {key} is not a string. Please check the formatting."
            )
        if not isinstance(value, str):
            raise click.ClickException(
                f"env_vars value {value} is not a string. Please check the formatting."
            )

    return env_vars


def _validate_and_modify_runtime_env(runtime_env: Dict):
    """Validates, then modifies runtime_env in place.

    Raises: click.ClickException if the runtime env doesn't pass validation.
    """

    if "conda" in runtime_env:
        conda_option = runtime_env["conda"]
        if not isinstance(conda_option, (str, dict)):
            raise click.ClickException(
                f"runtime_env['conda'] must be str or dict, got type({conda_option})."
            )
        runtime_env["conda"] = _validate_conda_option(conda_option)
    if "pip" in runtime_env:
        pip_option = runtime_env["pip"]
        if not isinstance(pip_option, (str, list)):
            raise click.ClickException(
                f"runtime_env['pip'] must be str or list, got type({pip_option})."
            )
        runtime_env["pip"] = _validate_pip_option(runtime_env["pip"])
    if "py_modules" in runtime_env:
        py_modules_option = runtime_env["py_modules"]
        if not isinstance(py_modules_option, list):
            raise click.ClickException(
                f"runtime_env['py_modules'] must be list, got type({py_modules_option})."
            )
        runtime_env["py_modules"] = _validate_py_modules(py_modules_option)
    if "upload_path" in runtime_env:
        upload_path_option = runtime_env["upload_path"]
        if not isinstance(upload_path_option, str):
            raise click.ClickException(
                f"runtime_env['upload_path'] must be str, got type({upload_path_option})."
            )
        runtime_env["upload_path"] = _validate_upload_path(upload_path_option)
    if "working_dir" in runtime_env:
        working_dir_option = runtime_env["working_dir"]
        if not isinstance(working_dir_option, str):
            raise click.ClickException(
                f"runtime_env['working_dir'] must be str, got type({working_dir_option})."
            )
        runtime_env["working_dir"] = _validate_working_dir(working_dir_option)
    _validate_working_dir_and_upload_path(
        runtime_env.get("working_dir"), runtime_env.get("upload_path")
    )
    if "env_vars" in runtime_env:
        env_vars_option = runtime_env["env_vars"]
        if not isinstance(env_vars_option, dict):
            raise click.ClickException(
                f"runtime_env['env_vars'] must be dict, got type({env_vars_option})."
            )
        runtime_env["env_vars"] = _validate_env_vars(env_vars_option)


def _get_project_id_from_id_or_name(
    project_id: Optional[str] = None, project_name: Optional[str] = None
) -> Optional[str]:
    """
    Get project id from PROJECT_NAME_ENV_VAR or `project_id` or `project_name`
    fields in the job config. Return None if no project was specified by the user,
    and the default project will selected later in the job submit code.
    """
    if project_id and project_name:
        raise click.ClickException(
            "Only one of `project_id` or `project` can be provided in the config file. "
        )
    project_name_env_var = os.environ.get(PROJECT_NAME_ENV_VAR)
    if project_name_env_var:
        # Get project from environment variable regardless of if is provided in config
        project_id = get_proj_id_from_name(project_name_env_var)
    elif project_name:
        project_id = get_proj_id_from_name(project_name)
    return project_id


class BaseHAJobConfig(BaseModel):
    """
    Base job configuration for both Production Jobs and Services.
    """

    name: Optional[str] = Field(
        None,
        description="Name of job to be submitted. Default will be used if not provided.",
    )
    description: Optional[str] = Field(
        None,
        description="Description of job to be submitted. Default will be used if not provided.",
    )
    runtime_env: Optional[Dict[str, Any]] = Field(
        None,
        description="A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.",
    )
    build_id: Optional[str] = Field(
        None,
        description="The id of the cluster env build. This id will determine the docker image your job is run on.",
    )
    cluster_env: Optional[str] = Field(
        None,
        description="The name of the cluster environment and build revision in format `my_cluster_env:1`.",
    )
    docker: Optional[str] = Field(None, description="Custom docker image name.")
    python_version: Optional[str] = Field(
        None, description="Python version for the custom docker image."
    )
    ray_version: Optional[str] = Field(
        None, description="Ray version for the custom docker image."
    )
    compute_config_id: Optional[str] = Field(
        None,
        description="The id of the compute configuration that you want to use. This id will specify the resources required for your job",
    )
    project_id: Optional[str] = Field(
        None,
        description="The id of the project you want to use. If not specified, and no project is inferred from the directory, no project will be used.",
    )
    workspace_id: Optional[str] = Field(
        None, description="The id of the workspace that this job is submitted from.",
    )
    project: Optional[str] = Field(
        None,
        description="The name of the project you want to use. If not specified, and no project is inferred from the directory, no project will be used.",
    )
    compute_config: Optional[Union[str, Dict[str, Any]]] = Field(
        None,
        description="The name of the compute configuration and version that you want to use in format `my_compute_config:1`. This will specify the resources required for your job."
        "This field also accepts a one-off config as a dictionary.",
    )
    cloud: Optional[str] = Field(
        None,
        description="The cloud name to specify a default compute configuration with. This will specify the resources required for your job",
    )
    max_retries: Optional[int] = Field(
        5,
        description="The number of retries this job will attempt on failure. Set to None to set infinite retries",
    )
    compute_config_version: Optional[int] = Field(
        None,
        description="The compute config version filter that you want to use. Only applicable if the compute_config was specified by name. "
        "Versions are positive integers. "
        "Deprecated behavior: Setting version to None is equivalent to setting version to '1'.",
    )

    @validator("compute_config_version")
    def optional_positive_int(cls, v):
        assert (
            v is None or v >= 1
        ), "compute_config_version must be a positive integer or None"
        return v

    class Config:
        extra = Extra.forbid

    @root_validator
    def fill_project_id(cls: Any, values: Any) -> Any:
        project_id, project_name = (
            values.get("project_id"),
            values.get("project"),
        )
        values["project_id"] = _get_project_id_from_id_or_name(project_id, project_name)
        return values

    @root_validator
    def fill_cluster_env_from_custom_docker(cls: Any, values: Any) -> Any:
        docker, python_version, ray_version, cluster_env, build_id = (
            values.get("docker"),
            values.get("python_version"),
            values.get("ray_version"),
            values.get("cluster_env"),
            values.get("build_id"),
        )
        if docker is not None:
            if cluster_env is not None:
                raise click.ClickException(
                    "`cluster_env` and `docker` cannot both be specified. Please only provide one"
                    "of these in the job config file."
                )
            if build_id is not None:
                raise click.ClickException(
                    "`build_id` and `docker` cannot both be specified. Please only provide one"
                    "of these in the job config file."
                )
            if python_version is None and ray_version is None:
                python_version, ray_version = extract_versions_from_image_name(docker)

            if python_version is None:
                raise click.ClickException(
                    "`python_version` should be specified when `docker` is used."
                )
            if ray_version is None:
                raise click.ClickException(
                    "`ray_version` should be specified when `docker` is used."
                )
            values["cluster_env"] = BYODInfo(
                docker_image_name=docker,
                python_version=python_version,
                ray_version=ray_version,
            ).encode()
        return values

    @root_validator
    def fill_build_id(cls: Any, values: Any) -> Any:
        build_id, cluster_env = (
            values.get("build_id"),
            values.get("cluster_env"),
        )
        if cluster_env and build_id:
            raise click.ClickException(
                "Only one of `cluster_env` or `build_id` can be provided in the config file. "
            )
        if cluster_env:
            build_id = get_build_from_cluster_env_identifier(cluster_env).id
            values["build_id"] = build_id
        elif not build_id:
            log.info(
                "No cluster environment provided, setting default based on local Python and Ray versions."
            )
            build_id = get_default_cluster_env_build().id
            values["build_id"] = build_id
        validate_successful_build(values["build_id"])
        return values

    @root_validator
    def fill_compute_config_id(cls: Any, values: Any) -> Any:
        compute_config_id, compute_config, cloud, compute_config_version = (
            values.get("compute_config_id"),
            values.get("compute_config"),
            values.get("cloud"),
            values.get("compute_config_version"),
        )
        project_id = values.get("project_id")
        if not project_id:
            project_id = _get_project_id_from_id_or_name(
                values.get("project_id"), values.get("project")
            )
        if bool(compute_config_id) + bool(compute_config) + bool(cloud) > 1:
            raise click.ClickException(
                "Only one of `compute_config_id`, `compute_config`, or `cloud` can be provided in the config file."
            )
        if compute_config and isinstance(compute_config, str):
            compute_config_id = get_cluster_compute_from_name(
                compute_config, version=compute_config_version
            ).id
        elif compute_config and isinstance(compute_config, dict):
            compute_config_id = register_compute_template(compute_config).id
        elif cloud:
            # Get default cluster compute for the specified cloud.
            compute_config_id = get_default_cluster_compute(
                cloud_name=cloud, project_id=None
            ).id
            log.info(
                f"Using default compute config for specified cloud {cloud}: {compute_config_id}."
            )
        elif not compute_config_id:
            parent_cloud_name = None
            if project_id:
                parent_cloud_id_and_name = get_parent_cloud_id_and_name_of_project(
                    project_id
                )
                if parent_cloud_id_and_name:
                    _, parent_cloud_name = parent_cloud_id_and_name
            # Get default cluster compute for the parent cloud if it exists or the default cloud default cloud.
            compute_config_id = get_default_cluster_compute(
                cloud_name=parent_cloud_name, project_id=None
            ).id
            msg_about_cloud = (
                f" for cloud {parent_cloud_name}" if parent_cloud_name else ""
            )
            log.info(
                f"No cloud or compute config specified, using the default{msg_about_cloud}: {compute_config_id}."
            )
        values["compute_config_id"] = compute_config_id

        return values

    @root_validator
    def validate_runtime_env(cls: Any, values: Any) -> Any:  # noqa: PLR0912
        runtime_env = values.get("runtime_env")
        if runtime_env is not None:
            _validate_and_modify_runtime_env(runtime_env)
        return values


class JobConfig(BaseHAJobConfig):
    """
    Job Config model for CLI. Validate and preprocess so `entrypoint`, `runtime_env_config`,
    `build_id`, `compute_config_id`, `max_retries` have the correct values to call
    `/api/v2/decorated_ha_jobs/create`.
    """

    entrypoint: str = Field(
        ...,
        description="A script that will be run to start your job. This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'",
    )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateFineTuningJobProductRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'training_file': 'str',
        'validation_file': 'str',
        'hyperparameters': 'CreateFineTuningHyperparameters',
        'model': 'str',
        'suffix': 'str',
        'cloud_id': 'str'
    }

    attribute_map = {
        'training_file': 'training_file',
        'validation_file': 'validation_file',
        'hyperparameters': 'hyperparameters',
        'model': 'model',
        'suffix': 'suffix',
        'cloud_id': 'cloud_id'
    }

    def __init__(self, training_file=None, validation_file=None, hyperparameters=None, model=None, suffix=None, cloud_id=None, local_vars_configuration=None):  # noqa: E501
        """CreateFineTuningJobProductRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._training_file = None
        self._validation_file = None
        self._hyperparameters = None
        self._model = None
        self._suffix = None
        self._cloud_id = None
        self.discriminator = None

        self.training_file = training_file
        if validation_file is not None:
            self.validation_file = validation_file
        if hyperparameters is not None:
            self.hyperparameters = hyperparameters
        self.model = model
        if suffix is not None:
            self.suffix = suffix
        self.cloud_id = cloud_id

    @property
    def training_file(self):
        """Gets the training_file of this CreateFineTuningJobProductRequest.  # noqa: E501

        ID of the .jsonl file used for training  # noqa: E501

        :return: The training_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._training_file

    @training_file.setter
    def training_file(self, training_file):
        """Sets the training_file of this CreateFineTuningJobProductRequest.

        ID of the .jsonl file used for training  # noqa: E501

        :param training_file: The training_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and training_file is None:  # noqa: E501
            raise ValueError("Invalid value for `training_file`, must not be `None`")  # noqa: E501

        self._training_file = training_file

    @property
    def validation_file(self):
        """Gets the validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :return: The validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._validation_file

    @validation_file.setter
    def validation_file(self, validation_file):
        """Sets the validation_file of this CreateFineTuningJobProductRequest.

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :param validation_file: The validation_file of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """

        self._validation_file = validation_file

    @property
    def hyperparameters(self):
        """Gets the hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501

        Hyperparameters used for fine-tuning.  # noqa: E501

        :return: The hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: CreateFineTuningHyperparameters
        """
        return self._hyperparameters

    @hyperparameters.setter
    def hyperparameters(self, hyperparameters):
        """Sets the hyperparameters of this CreateFineTuningJobProductRequest.

        Hyperparameters used for fine-tuning.  # noqa: E501

        :param hyperparameters: The hyperparameters of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: CreateFineTuningHyperparameters
        """

        self._hyperparameters = hyperparameters

    @property
    def model(self):
        """Gets the model of this CreateFineTuningJobProductRequest.  # noqa: E501

        A supported model used as a base for fine-tuning  # noqa: E501

        :return: The model of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._model

    @model.setter
    def model(self, model):
        """Sets the model of this CreateFineTuningJobProductRequest.

        A supported model used as a base for fine-tuning  # noqa: E501

        :param model: The model of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and model is None:  # noqa: E501
            raise ValueError("Invalid value for `model`, must not be `None`")  # noqa: E501

        self._model = model

    @property
    def suffix(self):
        """Gets the suffix of this CreateFineTuningJobProductRequest.  # noqa: E501

        A suffix (along with a hash) that will be appended to the model to create `fine_tuned_model`. Optional.  # noqa: E501

        :return: The suffix of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._suffix

    @suffix.setter
    def suffix(self, suffix):
        """Sets the suffix of this CreateFineTuningJobProductRequest.

        A suffix (along with a hash) that will be appended to the model to create `fine_tuned_model`. Optional.  # noqa: E501

        :param suffix: The suffix of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """

        self._suffix = suffix

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501


        :return: The cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CreateFineTuningJobProductRequest.


        :param cloud_id: The cloud_id of this CreateFineTuningJobProductRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateFineTuningJobProductRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateFineTuningJobProductRequest):
            return True

        return self.to_dict() != other.to_dict()



"""
Commands to interact with the Anyscale Session Commands API
"""

from typing import Optional

import click

from anyscale.authenticate import get_auth_api_client
from anyscale.formatters import common_formatter
from anyscale.sdk.anyscale_client import CreateSessionCommand


@click.group(
    "commands", help="Commands to interact with the Session Commands API.",
)
def session_commands() -> None:
    pass


@session_commands.command(name="list", short_help="Lists Session Commands.")
@click.argument("session-id", required=True)
@click.option(
    "--count", type=int, default=10, help="Number of commands to show. Defaults to 10"
)
@click.option(
    "--paging-token",
    required=False,
    help="Paging token used to fetch subsequent pages of commands.",
)
def list_session_commands(
    session_id: str, count: int, paging_token: Optional[str],
) -> None:
    """Lists commands that have been executed on Session with SESSION_ID"""

    api_client = get_auth_api_client().anyscale_api_client
    response = api_client.list_session_commands(
        session_id, count=count, paging_token=paging_token
    )

    print(common_formatter.prettify_json(response.to_dict()))


@session_commands.command(name="get", short_help="Get details about a Session Command.")
@click.argument("session-command-id", required=True)
def get_session_command(session_command_id: str) -> None:
    """Get details about the Session Command with id SESSION_COMMAND_ID"""

    api_client = get_auth_api_client().anyscale_api_client
    response = api_client.get_session_command(session_command_id)

    print(common_formatter.prettify_json(response.to_dict()))


@session_commands.command(name="kill", short_help="Kills a Session Command.")
@click.argument("session-command-id", required=True)
def kill_session_command(session_command_id: str) -> None:
    """Kills the Session Command with id SESSION_COMMAND_ID"""

    api_client = get_auth_api_client().anyscale_api_client
    response = api_client.kill_session_command(session_command_id)

    print(common_formatter.prettify_json(response.to_dict()))


@session_commands.command(
    name="create", short_help="Creates and executes a Session Command."
)
@click.argument("session-id", required=True)
@click.argument("shell-command", required=True)
def create_session_command(session_id: str, shell_command: str) -> None:
    """Creates and executes SHELL_COMMAND on the session with id SESSION_ID."""

    api_client = get_auth_api_client().anyscale_api_client
    create_data = CreateSessionCommand(
        session_id=session_id, shell_command=shell_command
    )
    response = api_client.create_session_command(create_data)

    print(common_formatter.prettify_json(response.to_dict()))



import os
from typing import Any, Dict, Optional, Tuple

import click
import yaml

import anyscale
from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.cloud import (
    get_cloud_id_and_name,
    get_last_used_cloud,
    get_organization_default_cloud,
)
import anyscale.conf
from anyscale.sdk.anyscale_client import (
    ComputeTemplateConfig,
    ComputeTemplateQuery,
)
from anyscale.utils.connect_helpers import find_project_id


# The cluster compute type. It can either be a string, eg my_template or a dict,
# eg, {"cloud_id": "id-123" ...}
CLUSTER_COMPUTE_DICT_TYPE = Dict[str, Any]


class ProjectBlock:
    """
    Class to determine which project ray.client uses. This class should never be
    instantiated directly. Instead call `get_project_block` to ensure a new
    ProjectBlock object is correctly created.
    """

    def __init__(  # noqa PLR0913
        self,
        project_dir: Optional[str] = None,
        project_name: Optional[str] = None,
        cloud_name: Optional[str] = None,
        cluster_compute_name: Optional[str] = None,
        cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE] = None,
        log_output: bool = True,
        cluster_compute_version: Optional[int] = None,
    ):
        """
        If project_dir is passed or exists in the directory structure: ensure the project_dir
            is correctly set up. Otherwise use the default project.
        Create an ProjectBlock object with the following properties that are useful
        externally: project_id, project_name, project_dir.

        Arguments:
            project_dir (Optional[str]): Project directory location. This doesn't need to contain
                a .anyscale.yaml file.
            project_name (Optional[str]): Project name if it is necessary to register the project
                with Anyscale. This will not be used if a registered project already exists
                in the project_dir.
            cloud_name (Optional[str]): This argument is only relevant if cloud isolation is enabled.
                If cloud_name is specified and no project arguements are specified, take the default
                project of the cloud.
        """
        self.log = BlockLogger(log_output=log_output)
        auth_api_client = get_auth_api_client(log_output=log_output)
        self.api_client = auth_api_client.api_client
        self.anyscale_api_client = auth_api_client.anyscale_api_client
        self.project_dir = (
            os.path.abspath(os.path.expanduser(project_dir)) if project_dir else None
        )
        self.cloud_name = cloud_name
        self.cluster_compute_name = cluster_compute_name
        self.cluster_compute_dict = cluster_compute_dict
        self.cluster_compute_version = cluster_compute_version

        self.block_label = "Project"
        self.log.open_block(self.block_label, block_title="Choosing a project")

        if self.project_dir is None and project_name is None:
            self.project_dir = anyscale.project.find_project_root(os.getcwd())

        parent_cloud_id = self._get_parent_cloud_id(
            self.cloud_name,
            self.cluster_compute_name,
            self.cluster_compute_dict,
            self.cluster_compute_version,
        )

        left_pad = " " * 2
        if self.project_dir:
            # TODO(nikita): Remove this logic when removing .anyscale.yaml in Q3 2022
            self.project_id, self.project_name = self._ensure_project_setup_at_dir(
                self.project_dir, project_name, parent_cloud_id
            )
            self.log.info(
                f"{left_pad}{'name:': <20}{self.project_name}",
                block_label=self.block_label,
            )
            self.log.info(
                f"{left_pad}{'project id:': <20}{self.project_id}",
                block_label=self.block_label,
            )
            self.log.info(
                f"{left_pad}{'project directory:': <20}{self.project_dir}",
                block_label=self.block_label,
            )
        elif project_name:
            self.project_id, self.project_name = self._create_or_get_project_from_name(
                project_name, parent_cloud_id
            )
            self.log.info(
                f"{left_pad}{'name:': <20}{self.project_name}",
                block_label=self.block_label,
            )
            self.log.info(
                f"{left_pad}{'project id:': <20}{self.project_id}",
                block_label=self.block_label,
            )
        else:
            self.project_id, self.project_name = self._get_default_project(
                parent_cloud_id
            )
        self.log.close_block(self.block_label)

    def _get_default_project(
        self, parent_cloud_id: Optional[str] = None
    ) -> Tuple[str, str]:
        """
        Get default project id and name.

        Returns:
        The project id and project name of the project being used.
        """
        default_project = self.anyscale_api_client.get_default_project(
            parent_cloud_id=parent_cloud_id
        ).result
        project_id = default_project.id
        project_name = default_project.name
        self.log.info(
            "No project defined. Continuing without a project.",
            block_label=self.block_label,
        )
        return project_id, project_name

    def _get_parent_cloud_id(
        self,
        cloud_name: Optional[str],
        cluster_compute_name: Optional[str],
        cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE],
        cluster_compute_version: Optional[int] = None,
    ) -> Optional[str]:
        parent_cloud_id = None
        if cloud_name:
            parent_cloud_id, _ = get_cloud_id_and_name(
                api_client=self.api_client, cloud_id=None, cloud_name=cloud_name,
            )
        elif cluster_compute_name:
            cluster_computes = self.api_client.search_compute_templates_api_v2_compute_templates_search_post(
                ComputeTemplateQuery(
                    orgwide=True,
                    name={"equals": cluster_compute_name},
                    include_anonymous=True,
                    version=cluster_compute_version,
                )
            ).results
            if len(cluster_computes) == 0:
                raise click.ClickException(
                    f"The cluster compute template {cluster_compute_name} does not exist."
                )
            parent_cloud_id = cluster_computes[0].config.cloud_id
        elif cluster_compute_dict:
            cluster_compute_config_obj = ComputeTemplateConfig(**cluster_compute_dict)
            parent_cloud_id = cluster_compute_config_obj.cloud_id
        else:
            # Get organization default cloud or last used cloud
            default_cloud_name = get_organization_default_cloud(self.api_client)
            if default_cloud_name:
                cloud_name = default_cloud_name
            else:
                cloud_name = get_last_used_cloud(None, self.anyscale_api_client)
            parent_cloud_id, _ = get_cloud_id_and_name(
                self.api_client, cloud_id=None, cloud_name=cloud_name
            )
        return parent_cloud_id

    def _ensure_project_setup_at_dir(
        self,
        project_dir: str,
        project_name: Optional[str],
        parent_cloud_id: Optional[str] = None,
    ) -> Tuple[str, str]:
        """
        Get or create an Anyscale project rooted at the given dir. If .anyscale.yaml
        exists in the given project dir and that project is registered with Anyscale,
        return information about that project. Otherwise, create a project with this
        project_dir and project_name (use default if not provided).

        Returns:
        The project id and project name of the project being used.
        """
        os.makedirs(project_dir, exist_ok=True)
        if project_name is None:
            project_name = os.path.basename(project_dir)

        # If the project yaml exists, assume we're already setup.
        project_yaml = os.path.join(project_dir, ".anyscale.yaml")
        if os.path.exists(project_yaml):
            # Validate format of project yaml and get project id
            proj_def = anyscale.project.ProjectDefinition(project_dir)
            project_id: Optional[str] = anyscale.project.get_project_id(proj_def.root)
            if not project_id:
                raise click.ClickException(
                    f"{project_yaml} is not correctly formatted. Please attach to a different "
                    "project using `anyscale project init`."
                )
            try:
                project_response = self.anyscale_api_client.get_project(project_id)
            except click.ClickException:
                raise click.ClickException(
                    f"Unable to get project with id {project_id} from Anyscale. Please attach to "
                    "a different project using `anyscale project init`."
                )
            self.log.info(
                f"Using the project defined in {project_yaml}:",
                block_label=self.block_label,
            )
            return project_id, project_response.result.name

        project_id = find_project_id(self.anyscale_api_client, project_name)
        if project_id is None:
            # Create a new project in the local directory with given name, because
            # project with this name doesn't exist yet.
            self.log.info(
                f"Creating new project named {BlockLogger.highlight(project_name)} for local dir {project_dir}.",
                block_label=self.block_label,
            )
            self.log.info(
                f"Using the project {BlockLogger.highlight(project_name)}:",
                block_label=self.block_label,
            )
            project_response = self.anyscale_api_client.create_project(
                {
                    "name": project_name,
                    "description": "Automatically created by Anyscale Connect",
                    "parent_cloud_id": parent_cloud_id,
                }
            )
            project_id = project_response.result.id
        else:
            # Project already exists with this name, yet directory doesn't contain
            # project yaml.
            self.log.info(
                f"Connecting local directory {project_dir} to project {BlockLogger.highlight(project_name)}.",
                block_label=self.block_label,
            )
            self.log.info(
                f"Using the project {BlockLogger.highlight(project_name)}:",
                block_label=self.block_label,
            )

        if not os.path.exists(project_yaml):
            with open(project_yaml, "w+") as f:
                f.write(yaml.dump({"project_id": project_id}))

        return project_id, project_name

    def _create_or_get_project_from_name(
        self, project_name: str, parent_cloud_id: Optional[str] = None
    ) -> Tuple[str, str]:
        """
        Get or create an Anyscale project not rooted in any directory
        (without .anyscale.yaml).

        Returns:
        The project id and project name of the project being used.
        """
        project_id = find_project_id(self.anyscale_api_client, project_name)
        if project_id is None:
            self.log.info(
                f"Creating new project named {BlockLogger.highlight(project_name)}.",
                block_label=self.block_label,
            )
            project_response = self.anyscale_api_client.create_project(
                {
                    "name": project_name,
                    "description": "Automatically created by Anyscale Connect",
                    "parent_cloud_id": parent_cloud_id,
                }
            )
            project_id = project_response.result.id
        self.log.info(
            f"Using the project {BlockLogger.highlight(project_name)}:",
            block_label=self.block_label,
        )
        return project_id, project_name


def create_project_block(
    project_dir: Optional[str] = None,
    project_name: Optional[str] = None,
    cloud_name: Optional[str] = None,
    cluster_compute_name: Optional[str] = None,
    cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE] = None,
    log_output: bool = True,
    cluster_compute_version: Optional[int] = None,
):
    """
    Function to create new ProjectBlock object. The ProjectBlock object
    is not a global variable an will be reinstantiated on each call to
    get_project_block.
    """
    return ProjectBlock(
        project_dir=project_dir,
        project_name=project_name,
        cloud_name=cloud_name,
        cluster_compute_name=cluster_compute_name,
        cluster_compute_dict=cluster_compute_dict,
        log_output=log_output,
        cluster_compute_version=cluster_compute_version,
    )



"""
Commands to interact with the Anyscale Session Operatons API
"""

import click

from anyscale.authenticate import get_auth_api_client
from anyscale.formatters import common_formatter


@click.group(
    "session_operations", help="Commands to interact with the Session Operations API.",
)
def session_operations_commands() -> None:
    pass


@session_operations_commands.command(
    name="get", short_help="Retrieves a Session Operation."
)
@click.argument("session_operation_id", required=True)
def get_session_operation(session_operation_id: str,) -> None:
    """Retrieves a Session Operation.
    """

    api_client = get_auth_api_client().anyscale_api_client
    response = api_client.get_session_operation(session_operation_id)
    print(common_formatter.prettify_json(response.to_dict()))



from unittest.mock import Mock

import pytest

from anyscale.client.openapi_client.models import (
    ListMachinesResponse,
    MachineAllocationState,
    MachineConnectionState,
    MachineInfo,
)
from frontend.cli.anyscale.controllers.machines_controller import MachinesController


@pytest.mark.parametrize("cloud_id", ["cld_123"])
def test_list_machines(mock_auth_api_client, cloud_id: str) -> None:
    response = ListMachinesResponse(
        machines=[
            MachineInfo(
                machine_id="m-123",
                hostname="dummy",
                machine_shape="lambda-a100-80g",
                connection_state=MachineConnectionState.CONNECTED,
                allocation_state=MachineAllocationState.AVAILABLE,
                cluster_id="",
                node_id="",
            )
        ]
    )

    api_response = Mock()
    api_response.result = response

    machines_controller = MachinesController()
    machines_controller.api_client.list_machines_api_v2_machines_get = Mock(
        return_value=api_response
    )
    output = machines_controller.list_machines(cloud_id=cloud_id)
    assert output == response



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class AicaModel(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'emoji': 'str',
        'description': 'str',
        'base_canonical_model_id': 'str',
        'min_aviary_version': 'str',
        'max_aviary_version': 'str',
        'model_object_storage_uri': 'str',
        'is_model_stored_in_anyscale_account': 'bool',
        'generation_configuration_json': 'object',
        'created_at': 'datetime',
        'cloud_id': 'str'
    }

    attribute_map = {
        'id': 'id',
        'emoji': 'emoji',
        'description': 'description',
        'base_canonical_model_id': 'base_canonical_model_id',
        'min_aviary_version': 'min_aviary_version',
        'max_aviary_version': 'max_aviary_version',
        'model_object_storage_uri': 'model_object_storage_uri',
        'is_model_stored_in_anyscale_account': 'is_model_stored_in_anyscale_account',
        'generation_configuration_json': 'generation_configuration_json',
        'created_at': 'created_at',
        'cloud_id': 'cloud_id'
    }

    def __init__(self, id=None, emoji=None, description=None, base_canonical_model_id=None, min_aviary_version=None, max_aviary_version=None, model_object_storage_uri=None, is_model_stored_in_anyscale_account=None, generation_configuration_json=None, created_at=None, cloud_id=None, local_vars_configuration=None):  # noqa: E501
        """AicaModel - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._emoji = None
        self._description = None
        self._base_canonical_model_id = None
        self._min_aviary_version = None
        self._max_aviary_version = None
        self._model_object_storage_uri = None
        self._is_model_stored_in_anyscale_account = None
        self._generation_configuration_json = None
        self._created_at = None
        self._cloud_id = None
        self.discriminator = None

        self.id = id
        if emoji is not None:
            self.emoji = emoji
        if description is not None:
            self.description = description
        if base_canonical_model_id is not None:
            self.base_canonical_model_id = base_canonical_model_id
        if min_aviary_version is not None:
            self.min_aviary_version = min_aviary_version
        if max_aviary_version is not None:
            self.max_aviary_version = max_aviary_version
        if model_object_storage_uri is not None:
            self.model_object_storage_uri = model_object_storage_uri
        if is_model_stored_in_anyscale_account is not None:
            self.is_model_stored_in_anyscale_account = is_model_stored_in_anyscale_account
        if generation_configuration_json is not None:
            self.generation_configuration_json = generation_configuration_json
        self.created_at = created_at
        if cloud_id is not None:
            self.cloud_id = cloud_id

    @property
    def id(self):
        """Gets the id of this AicaModel.  # noqa: E501

        Name of the model to be created.  # noqa: E501

        :return: The id of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this AicaModel.

        Name of the model to be created.  # noqa: E501

        :param id: The id of this AicaModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def emoji(self):
        """Gets the emoji of this AicaModel.  # noqa: E501

        The emoji of the model  # noqa: E501

        :return: The emoji of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._emoji

    @emoji.setter
    def emoji(self, emoji):
        """Sets the emoji of this AicaModel.

        The emoji of the model  # noqa: E501

        :param emoji: The emoji of this AicaModel.  # noqa: E501
        :type: str
        """

        self._emoji = emoji

    @property
    def description(self):
        """Gets the description of this AicaModel.  # noqa: E501

        The description of the model  # noqa: E501

        :return: The description of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this AicaModel.

        The description of the model  # noqa: E501

        :param description: The description of this AicaModel.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def base_canonical_model_id(self):
        """Gets the base_canonical_model_id of this AicaModel.  # noqa: E501

        The name of the base model. This field is present only if the model is finetuned.  # noqa: E501

        :return: The base_canonical_model_id of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._base_canonical_model_id

    @base_canonical_model_id.setter
    def base_canonical_model_id(self, base_canonical_model_id):
        """Sets the base_canonical_model_id of this AicaModel.

        The name of the base model. This field is present only if the model is finetuned.  # noqa: E501

        :param base_canonical_model_id: The base_canonical_model_id of this AicaModel.  # noqa: E501
        :type: str
        """

        self._base_canonical_model_id = base_canonical_model_id

    @property
    def min_aviary_version(self):
        """Gets the min_aviary_version of this AicaModel.  # noqa: E501


        :return: The min_aviary_version of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._min_aviary_version

    @min_aviary_version.setter
    def min_aviary_version(self, min_aviary_version):
        """Sets the min_aviary_version of this AicaModel.


        :param min_aviary_version: The min_aviary_version of this AicaModel.  # noqa: E501
        :type: str
        """

        self._min_aviary_version = min_aviary_version

    @property
    def max_aviary_version(self):
        """Gets the max_aviary_version of this AicaModel.  # noqa: E501


        :return: The max_aviary_version of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._max_aviary_version

    @max_aviary_version.setter
    def max_aviary_version(self, max_aviary_version):
        """Sets the max_aviary_version of this AicaModel.


        :param max_aviary_version: The max_aviary_version of this AicaModel.  # noqa: E501
        :type: str
        """

        self._max_aviary_version = max_aviary_version

    @property
    def model_object_storage_uri(self):
        """Gets the model_object_storage_uri of this AicaModel.  # noqa: E501

        The URI of the model object in the object storage.  # noqa: E501

        :return: The model_object_storage_uri of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._model_object_storage_uri

    @model_object_storage_uri.setter
    def model_object_storage_uri(self, model_object_storage_uri):
        """Sets the model_object_storage_uri of this AicaModel.

        The URI of the model object in the object storage.  # noqa: E501

        :param model_object_storage_uri: The model_object_storage_uri of this AicaModel.  # noqa: E501
        :type: str
        """

        self._model_object_storage_uri = model_object_storage_uri

    @property
    def is_model_stored_in_anyscale_account(self):
        """Gets the is_model_stored_in_anyscale_account of this AicaModel.  # noqa: E501

        Whether the model is stored in the anyscale account or in the customer's account.  # noqa: E501

        :return: The is_model_stored_in_anyscale_account of this AicaModel.  # noqa: E501
        :rtype: bool
        """
        return self._is_model_stored_in_anyscale_account

    @is_model_stored_in_anyscale_account.setter
    def is_model_stored_in_anyscale_account(self, is_model_stored_in_anyscale_account):
        """Sets the is_model_stored_in_anyscale_account of this AicaModel.

        Whether the model is stored in the anyscale account or in the customer's account.  # noqa: E501

        :param is_model_stored_in_anyscale_account: The is_model_stored_in_anyscale_account of this AicaModel.  # noqa: E501
        :type: bool
        """

        self._is_model_stored_in_anyscale_account = is_model_stored_in_anyscale_account

    @property
    def generation_configuration_json(self):
        """Gets the generation_configuration_json of this AicaModel.  # noqa: E501

        The configuration for the generation section. This field is present only if the model is finetuned.  # noqa: E501

        :return: The generation_configuration_json of this AicaModel.  # noqa: E501
        :rtype: object
        """
        return self._generation_configuration_json

    @generation_configuration_json.setter
    def generation_configuration_json(self, generation_configuration_json):
        """Sets the generation_configuration_json of this AicaModel.

        The configuration for the generation section. This field is present only if the model is finetuned.  # noqa: E501

        :param generation_configuration_json: The generation_configuration_json of this AicaModel.  # noqa: E501
        :type: object
        """

        self._generation_configuration_json = generation_configuration_json

    @property
    def created_at(self):
        """Gets the created_at of this AicaModel.  # noqa: E501

        Time at which the model was created.  # noqa: E501

        :return: The created_at of this AicaModel.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this AicaModel.

        Time at which the model was created.  # noqa: E501

        :param created_at: The created_at of this AicaModel.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def cloud_id(self):
        """Gets the cloud_id of this AicaModel.  # noqa: E501

        The cloud id for the model. If None, this model is a base model.  # noqa: E501

        :return: The cloud_id of this AicaModel.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this AicaModel.

        The cloud id for the model. If None, this model is a base model.  # noqa: E501

        :param cloud_id: The cloud_id of this AicaModel.  # noqa: E501
        :type: str
        """

        self._cloud_id = cloud_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AicaModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AicaModel):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime, timezone
import json
from typing import Dict, List, Optional
from unittest.mock import MagicMock, Mock, patch

import boto3
from botocore.exceptions import ClientError, NoCredentialsError
from click import ClickException
from moto import mock_cloudformation, mock_iam
import pytest

from anyscale.aws_iam_policies import ANYSCALE_IAM_POLICIES, AnyscaleIAMPolicy
from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models import Cloud
from anyscale.utils.cloud_update_utils import (
    add_missing_parameters_to_template_body,
    append_statements_to_customer_drifts_policy,
    AWSCloudformationHandler,
    CUSTOMER_DRIFTS_POLICY_NAME,
    extract_cross_account_iam_role_drift,
    extract_drifted_statements,
    generate_drifted_statements_to_append,
    generate_updated_parameters_and_template,
    get_all_sids,
    get_all_sids_from_policy,
    get_or_create_memorydb,
    get_sids_to_remove,
    is_template_policy_documents_up_to_date,
    merge_parameters,
    PropertyDifference,
    try_delete_customer_drifts_policy,
    update_iam_role,
    update_template_with_memorydb,
    validate_stack_version,
)


def get_mock_cloud():
    mock_cloud = Cloud(
        id="cloud_id_1",
        name="cloud_name_1",
        provider="AWS",
        region="us-west-2",
        credentials="credentials",
        creator_id="creator_id",
        type="PUBLIC",
        created_at=datetime.now(timezone.utc),
        config="",
        state="ACTIVE",
        is_bring_your_own_resource=False,
    )
    return mock_cloud


def generate_mock_statement(sid="IAM") -> Dict:
    statement: Dict = {
        "Sid": sid,
        "Effect": "Allow",
        "Action": ["iam:PassRole", "iam:GetInstanceProfile"],
        "Resource": "*",
    }
    return statement


@pytest.mark.parametrize(
    ("mock_parameter_list", "expected_result"),
    [
        pytest.param([], False, id="no_parameters"),
        pytest.param(
            [
                {
                    "ParameterKey": policy.parameter_key,
                    "ParameterValue": policy.policy_document,
                }
                for policy in ANYSCALE_IAM_POLICIES
            ],
            True,
            id="up_to_date",
        ),
        pytest.param(
            [
                {
                    "ParameterKey": policy.parameter_key,
                    "ParameterValue": policy.policy_document + "extra",
                }
                for policy in ANYSCALE_IAM_POLICIES
            ],
            False,
            id="not_up_to_date",
        ),
    ],
)
def test_is_template_policy_documents_up_to_date(mock_parameter_list, expected_result):
    assert (
        is_template_policy_documents_up_to_date(mock_parameter_list) == expected_result
    )


def _get_cloudformation_template_and_parameters():
    cfn_template_body = """Description: This template creates the resources necessary for an anyscale cloud.
Transform: AWS::LanguageExtensions
Parameters:
  AnyscaleCrossAccountIAMRoleName:
    Description: Name of the cross account IAM role.
    Type: String

  AnyscaleCrossAccountIAMPolicySteadyState:
    Description: Stead state IAM policy document
    Type: String

  AnyscaleCrossAccountIAMPolicyServiceSteadyState:
    Description: Stead state IAM policy document for services
    Type: String

  AnyscaleCrossAccountIAMPolicyInitialRun:
    Description: Initial run IAM policy document
    Type: String

Resources:
  customerRole:
    Type: 'AWS::IAM::Role'
    Properties:
      RoleName: !Ref AnyscaleCrossAccountIAMRoleName
      AssumeRolePolicyDocument:
        Statement:
          - Action: 'sts:AssumeRole'
            Effect: Allow
            Principal:
              AWS: 525325868955
            Sid: 'AnyscaleControlPlaneAssumeRole'
        Version: 2012-10-17
      Path: /

  IAMPermissionEC2SteadyState:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicySteadyState
      PolicyName: Anyscale_IAM_Policy_Steady_State
      Roles:
        - !Ref customerRole

  IAMPermissionServiceSteadyState:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicyServiceSteadyState
      PolicyName: Anyscale_IAM_Policy_Service_Steady_State
      Roles:
        - !Ref customerRole

  IAMPermissionEC2InitialRun:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicyInitialRun
      PolicyName: Anyscale_IAM_Policy_Initial_Setup
      Roles:
        - !Ref customerRole
"""
    mock_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "IAM",
                "Effect": "Allow",
                "Action": ["iam:PassRole", "iam:GetInstanceProfile"],
                "Resource": "*",
            },
        ],
    }
    parameters = [
        {
            "ParameterKey": "AnyscaleCrossAccountIAMRoleName",
            "ParameterValue": "anyscale-iam-role",
        },
        {
            "ParameterKey": "AnyscaleCrossAccountIAMPolicySteadyState",
            "ParameterValue": json.dumps(mock_policy),
        },
        {
            "ParameterKey": "AnyscaleCrossAccountIAMPolicyServiceSteadyState",
            "ParameterValue": json.dumps(mock_policy),
        },
        {
            "ParameterKey": "AnyscaleCrossAccountIAMPolicyInitialRun",
            "ParameterValue": json.dumps(mock_policy),
        },
    ]
    return cfn_template_body, parameters


@pytest.mark.parametrize(
    ("existing_parameters", "parameters_to_update", "expected_merged_parameters"),
    [
        pytest.param(
            [
                {"ParameterKey": "key1", "ParameterValue": "value1"},
                {"ParameterKey": "key2", "ParameterValue": "value2"},
            ],
            [
                {"ParameterKey": "key1", "ParameterValue": "value1"},
                {"ParameterKey": "key2", "ParameterValue": "newvalue2"},
            ],
            [
                {"ParameterKey": "key1", "ParameterValue": "value1"},
                {"ParameterKey": "key2", "ParameterValue": "newvalue2"},
            ],
        ),
        pytest.param(
            [
                {"ParameterKey": "key1", "ParameterValue": "value1"},
                {"ParameterKey": "key2", "ParameterValue": "value2"},
                {"ParameterKey": "key4", "ParameterValue": "value4"},
            ],
            [
                {"ParameterKey": "key1", "ParameterValue": "newvalue1"},
                {"ParameterKey": "key3", "ParameterValue": "newvalue3"},
            ],
            [
                {"ParameterKey": "key1", "ParameterValue": "newvalue1"},
                {"ParameterKey": "key2", "ParameterValue": "value2"},
                {"ParameterKey": "key4", "ParameterValue": "value4"},
                {"ParameterKey": "key3", "ParameterValue": "newvalue3"},
            ],
            id="missing_parameters",
        ),
    ],
)
def test_merge_parameters(
    existing_parameters: List[Dict],
    parameters_to_update: List[Dict],
    expected_merged_parameters: List[str],
):
    assert (
        merge_parameters(existing_parameters, parameters_to_update)
        == expected_merged_parameters
    )


def test_add_missing_parameters_to_template_body():
    mock_template_body, mock_parameters = _get_cloudformation_template_and_parameters()
    mock_new_policy = AnyscaleIAMPolicy(
        parameter_key="NewParameter",
        parameter_description="NewParameter description",
        resource_logical_id="NewParameterResource",
        policy_name="New_Parameter_Policy",
        policy_document='{"Version": "2012-10-17", "Statement": []}',
    )
    mock_parameters.append(
        {
            "ParameterKey": mock_new_policy.parameter_key,
            "ParameterValue": mock_new_policy.policy_document,
        }
    )
    with patch.multiple(
        "anyscale.utils.cloud_update_utils", ANYSCALE_IAM_POLICIES=[mock_new_policy]
    ):
        modified_template_body = add_missing_parameters_to_template_body(
            mock_template_body,
            ["NewParameter", "AnyscaleCLIVersion", "MemoryDBRedisPort"],
        )
        assert (
            modified_template_body
            == f"""Description: This template creates the resources necessary for an anyscale cloud.
Transform: AWS::LanguageExtensions
Parameters:
  {mock_new_policy.parameter_key}:
    Description: {mock_new_policy.parameter_description}
    Type: String

  AnyscaleCLIVersion:
    Description: Anyscale CLI version
    Type: String

  MemoryDBRedisPort:
    Description: Port for MemoryDB Redis
    Type: String
    Default: \"6379\"

  AnyscaleCrossAccountIAMRoleName:
    Description: Name of the cross account IAM role.
    Type: String

  AnyscaleCrossAccountIAMPolicySteadyState:
    Description: Stead state IAM policy document
    Type: String

  AnyscaleCrossAccountIAMPolicyServiceSteadyState:
    Description: Stead state IAM policy document for services
    Type: String

  AnyscaleCrossAccountIAMPolicyInitialRun:
    Description: Initial run IAM policy document
    Type: String

Resources:
  {mock_new_policy.resource_logical_id}:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref {mock_new_policy.parameter_key}
      PolicyName: {mock_new_policy.policy_name}
      Roles:
        - !Ref customerRole

  customerRole:
    Type: 'AWS::IAM::Role'
    Properties:
      RoleName: !Ref AnyscaleCrossAccountIAMRoleName
      AssumeRolePolicyDocument:
        Statement:
          - Action: 'sts:AssumeRole'
            Effect: Allow
            Principal:
              AWS: 525325868955
            Sid: 'AnyscaleControlPlaneAssumeRole'
        Version: 2012-10-17
      Path: /

  IAMPermissionEC2SteadyState:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicySteadyState
      PolicyName: Anyscale_IAM_Policy_Steady_State
      Roles:
        - !Ref customerRole

  IAMPermissionServiceSteadyState:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicyServiceSteadyState
      PolicyName: Anyscale_IAM_Policy_Service_Steady_State
      Roles:
        - !Ref customerRole

  IAMPermissionEC2InitialRun:
    Type: AWS::IAM::Policy
    Properties:
      PolicyDocument: !Ref AnyscaleCrossAccountIAMPolicyInitialRun
      PolicyName: Anyscale_IAM_Policy_Initial_Setup
      Roles:
        - !Ref customerRole
"""
        )


@pytest.mark.parametrize(
    ("mock_drift", "has_drift"),
    [
        pytest.param(
            {
                "ResourceType": "AWS::IAM::Role",
                "LogicalResourceId": "customerRole",
                "StackResourceDriftStatus": "MODIFIED",
            },
            True,
            id="has_drift",
        ),
        pytest.param(MagicMock(), False, id="no_drift"),
    ],
)
def test_extract_cross_account_iam_role_drift(mock_drift, has_drift):
    assert (extract_cross_account_iam_role_drift([mock_drift]) is not None) == has_drift


@pytest.mark.parametrize(
    ("all_have_sid"),
    [
        pytest.param(True, id="all_have_sid"),
        pytest.param(False, id="not_all_have_sid"),
    ],
)
def test_get_all_sids_from_policy(all_have_sid: bool):
    policy: Dict = {
        "PolicyDocument": {
            "Version": "2012-10-17",
            "Statement": [generate_mock_statement(),],
        },
        "PolicyName": "mock_policy_name",
    }
    if not all_have_sid:
        statements = policy["PolicyDocument"]["Statement"]
        statements[0].pop("Sid")
        with pytest.raises(ClickException) as e:
            get_all_sids_from_policy(policy)
        e.match("doesn't have a Sid")
    else:
        assert get_all_sids_from_policy(policy) == ["IAM"]


@pytest.mark.parametrize(
    ("no_statement_number", "policy_drift_unexpected"),
    [
        pytest.param(True, False, id="no_statement_number"),
        pytest.param(False, True, id="policy_drift_unexpected"),
        pytest.param(False, False, id="happy_path"),
    ],
)
def test_get_sids_to_remove(no_statement_number: bool, policy_drift_unexpected: bool):
    diffs: List[Dict] = [
        # Not a policy drift
        {
            "PropertyPath": "/Something/Else",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "NOT_EQUAL",
        },
        # Not a policy we care about
        {
            "PropertyPath": "/Policies/0",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "NOT_EQUAL",
        },
        # REMOVE
        {
            "PropertyPath": "/Policies/1/PolicyDocument/Statement/0",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "REMOVE",
        },
        # ADD
        {
            "PropertyPath": "/Policies/1/PolicyDocument/Statement/10",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "ADD",
        },
        # statement_number longer than number of statements
        {
            "PropertyPath": "/Policies/1/PolicyDocument/Statement/100/Action/4",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "NOT_EQUAL",
        },
        # sid to remove
        {
            "PropertyPath": "/Policies/1/PolicyDocument/Statement/1/Action/1",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "NOT_EQUAL",
        },
        # sid already removed
        {
            "PropertyPath": "/Policies/1/PolicyDocument/Statement/1/Action/0",
            "ExpectedValue": "sth",
            "ActualValue": "sth",
            "DifferenceType": "NOT_EQUAL",
        },
    ]
    if no_statement_number:
        diffs.append(
            {
                "PropertyPath": "/Policies/1/PolicyDocument/nostatementnumber",
                "ExpectedValue": "sth",
                "ActualValue": "sth",
                "DifferenceType": "NOT_EQUAL",
            }
        )
    elif policy_drift_unexpected:
        diffs.append(
            {
                "PropertyPath": "/Policies/1/PolicyDocument/Statement/10",
                "ExpectedValue": "sth",
                "ActualValue": "sth",
                "DifferenceType": "NOT_EQUAL",  # shouldn't be not_equal
            }
        )
    mock_policy_name = ANYSCALE_IAM_POLICIES[0].policy_name

    expected_policies = [
        {
            "PolicyName": "AnotherPolicy",
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [generate_mock_statement(),],
            },
        },
        {
            "PolicyName": mock_policy_name,
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [Mock(), generate_mock_statement(),],
            },
        },
    ]

    if no_statement_number or policy_drift_unexpected:
        with pytest.raises(ClickException) as e:
            get_sids_to_remove(diffs, expected_policies)
        e.match("cannot be resolved")
    else:
        sids = get_sids_to_remove(diffs, expected_policies)
        assert len(sids[mock_policy_name]) == 1


def test_generate_drifted_statements_to_append():
    statement_with_no_sid = generate_mock_statement()
    statement_with_no_sid.pop("Sid")
    mock_policy_name = "mock_policy_name"
    actual_policies = [
        {
            "PolicyName": mock_policy_name,
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    generate_mock_statement(),
                    generate_mock_statement("AnotherSid"),
                    statement_with_no_sid,
                ],
            },
        },
        {
            "PolicyName": "AnotherPolicy",
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [generate_mock_statement(),],
            },
        },
    ]
    undrifted_sid = {mock_policy_name: {"IAM"}}
    drifted_statements = generate_drifted_statements_to_append(
        actual_policies, undrifted_sid
    )
    assert len(drifted_statements) == 2
    for statement in drifted_statements:
        assert "Drifted" in statement["Sid"]


@pytest.mark.parametrize("invalid_version", [True, False])
def test_extract_drifted_statements(invalid_version: bool):
    mock_policy_name = "mock_policy_name"
    mock_actual_policies = "mock"
    mock_drift: Dict = {
        "ExpectedProperties": json.dumps(
            {
                "Policies": [
                    {
                        "PolicyName": mock_policy_name,
                        "PolicyDocument": {
                            "Version": "wrong_version"
                            if invalid_version
                            else "2012-10-17",
                            "Statement": [
                                generate_mock_statement(),
                                generate_mock_statement("AnotherSid"),
                            ],
                        },
                    },
                    {
                        "PolicyName": "AnotherPolicy",
                        "PolicyDocument": {
                            "Version": "wrong_version",  # shouldn't raise exception
                            "Statement": [generate_mock_statement(),],
                        },
                    },
                ]
            }
        ),
        "PropertyDifferences": Mock(),
        "ActualProperties": json.dumps({"Policies": mock_actual_policies,}),
    }
    mock_get_all_sids_from_policy = Mock(return_value=["IAM", "AnotherSid"])
    mock_get_sids_to_remove = Mock(
        return_value={mock_policy_name: {"IAM", "AnotherActualSid"}}
    )
    mock_generate_drifted_statements_to_append = Mock()
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        ANYSCALE_IAM_POLICIES=[MagicMock(policy_name=mock_policy_name)],
        get_sids_to_remove=mock_get_sids_to_remove,
        generate_drifted_statements_to_append=mock_generate_drifted_statements_to_append,
        get_all_sids_from_policy=mock_get_all_sids_from_policy,
    ):
        if invalid_version:
            with pytest.raises(ClickException) as e:
                extract_drifted_statements(mock_drift)
            e.match("Unexpected policy version")
        else:
            extract_drifted_statements(mock_drift)
            mock_get_all_sids_from_policy.assert_called_once()
            mock_get_sids_to_remove.assert_called_once()
            mock_generate_drifted_statements_to_append.assert_called_once_with(
                mock_actual_policies, {mock_policy_name: {"AnotherSid"}}
            )


@mock_iam
@pytest.mark.parametrize("policy_exists", [True, False])
def test_append_statements_to_customer_drifts_policy(policy_exists: bool):
    mock_region = "us-west-2"
    mock_role_name = "mock_role_name"
    statements = [generate_mock_statement("DriftedSid")]

    # Create the role
    boto3.client("iam", region_name=mock_region).create_role(
        Path="/",
        RoleName=mock_role_name,
        AssumeRolePolicyDocument=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": "mock"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            }
        ),
    )
    if policy_exists:
        boto3.client("iam", region_name=mock_region).put_role_policy(
            RoleName=mock_role_name,
            PolicyName=CUSTOMER_DRIFTS_POLICY_NAME,
            PolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [generate_mock_statement("MockSid")],
                }
            ),
        )

    append_statements_to_customer_drifts_policy(mock_region, mock_role_name, statements)

    # Check that the policy has been appended
    policy = boto3.client("iam", region_name=mock_region).get_role_policy(
        RoleName=mock_role_name, PolicyName=CUSTOMER_DRIFTS_POLICY_NAME
    )
    assert "DriftedSid" in [s["Sid"] for s in policy["PolicyDocument"]["Statement"]]


@pytest.mark.parametrize("invalid_version", [True, False])
def test_get_all_sids(invalid_version: bool):
    mock_policy_name = "mock_policy_name"
    mock_expected_policies = [
        {
            "PolicyName": mock_policy_name,
            "PolicyDocument": {
                "Version": "wrong_version" if invalid_version else "2012-10-17",
                "Statement": [
                    generate_mock_statement("OneSid"),
                    generate_mock_statement("AnotherSid"),
                ],
            },
        },
        {
            "PolicyName": "AnotherPolicy",
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [generate_mock_statement(),],
            },
        },
    ]
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        ANYSCALE_IAM_POLICIES=[MagicMock(policy_name=mock_policy_name)],
    ):
        if invalid_version:
            with pytest.raises(ClickException) as e:
                get_all_sids(mock_expected_policies)
            e.match("Unexpected policy version")
        else:
            sids = get_all_sids(mock_expected_policies)
            assert len(sids) == 1
            assert mock_policy_name in sids
            assert len(sids[mock_policy_name]) == 2
            assert "OneSid" in sids[mock_policy_name]
            assert "AnotherSid" in sids[mock_policy_name]


@pytest.mark.parametrize(
    ("path", "expected_result"),
    [
        pytest.param(
            "/Policies/1024/PolicyDocument/Statement/0/Action/0",
            1024,
            id="is_policy_drift",
        ),
        pytest.param("/Something/Else", None, id="is_not_policy_drift"),
    ],
)
def test_get_policy_number(path: str, expected_result: Optional[int]):
    mock_diff = PropertyDifference(
        DifferenceType="NOT_EQUAL",
        PropertyPath=path,
        ExpectedValue="iam:PassRole",
        ActualValue="iam:PassRole",
    )
    assert mock_diff.get_policy_number() == expected_result


@pytest.mark.parametrize(
    ("path", "expected_result"),
    [
        pytest.param(
            "/Policies/1024/PolicyDocument/Statement/100/Action/20",
            100,
            id="is_action_drift",
        ),
        pytest.param("/Something/Else", None, id="is_not_policy_drift"),
        pytest.param("/Policies/3", None, id="is_not_statement_drift"),
        pytest.param(
            "/Policies/3/PolicyDocument/Statement/1024", 1024, id="is_statement_drift"
        ),
    ],
)
def test_get_statement_number(path: str, expected_result: Optional[int]):
    mock_diff = PropertyDifference(
        DifferenceType="NOT_EQUAL",
        PropertyPath=path,
        ExpectedValue="iam:PassRole",
        ActualValue="iam:PassRole",
    )
    assert mock_diff.get_statement_number() == expected_result


@pytest.mark.parametrize(
    ("path", "expected_result", "invalid"),
    [
        pytest.param(
            "/Policies/1024/PolicyDocument/Statement/100/Action/20",
            False,
            True,
            id="is_action_drift",
        ),
        pytest.param("/Something/Else", False, False, id="is_not_policy_drift"),
        pytest.param("/Policies/3", False, False, id="is_policy_drift"),
        pytest.param(
            "/Policies/3/PolicyDocument/Statement/1024",
            True,
            False,
            id="is_statement_drift",
        ),
        pytest.param(
            "/Policies/3/PolicyDocument/Statement/1024", True, True, id="invalid"
        ),
    ],
)
def test_is_add_or_remove_statement(path: str, expected_result: bool, invalid: bool):
    mock_diff = PropertyDifference(
        DifferenceType="ADD" if not invalid else "NOT_EQUAL",
        PropertyPath=path,
        ExpectedValue="iam:PassRole",
        ActualValue="iam:PassRole",
    )
    if expected_result and invalid:
        with pytest.raises(ClickException) as e:
            mock_diff.is_add_or_remove_statement()
        e.match("cannot be resolved")
    else:
        assert mock_diff.is_add_or_remove_statement() == expected_result


@mock_iam
@pytest.mark.parametrize(
    ("policy_exists", "delete_failed"),
    [
        pytest.param(True, False, id="policy_exists"),
        pytest.param(False, False, id="policy_does_not_exist"),
        pytest.param(False, True, id="delete_failed"),
    ],
)
def test_try_delete_customer_drifts_policy(policy_exists: bool, delete_failed: bool):
    cloud = get_mock_cloud()
    iam_client = boto3.client("iam", region_name=cloud.region)
    mock_role_name = "mock_role_name"
    cloud.credentials = "arn:aws:iam::123456789012:role/" + mock_role_name

    iam_client.create_role(
        Path="/",
        RoleName=mock_role_name,
        AssumeRolePolicyDocument=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": "mock"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            }
        ),
    )

    if policy_exists:
        iam_client.put_role_policy(
            RoleName=mock_role_name,
            PolicyName=CUSTOMER_DRIFTS_POLICY_NAME,
            PolicyDocument=json.dumps(
                {"Version": "2012-10-17", "Statement": [generate_mock_statement()],}
            ),
        )

    if delete_failed:
        iam_client.delete_role_policy = Mock(
            side_effect=ClientError(
                {"Error": {"Code": "Mock", "Message": "Mock Error",}},
                "delete_role_policy",
            )
        )
        with pytest.raises(ClickException) as e, patch.multiple(
            "anyscale.utils.cloud_update_utils", _client=Mock(return_value=iam_client)
        ):
            try_delete_customer_drifts_policy(cloud)
        e.match("Failed to delete inline policy")
    else:
        try_delete_customer_drifts_policy(cloud)
        assert (
            CUSTOMER_DRIFTS_POLICY_NAME
            not in iam_client.list_role_policies(RoleName=mock_role_name)["PolicyNames"]
        )


@pytest.mark.parametrize(
    (
        "describe_stack_client_error",
        "describe_stack_no_credentials",
        "stack_num",
        "stack_status_invalid",
        "expected_error",
    ),
    [
        pytest.param(False, False, 1, False, False, id="happy_path"),
        pytest.param(True, False, 1, False, True, id="describe_stack_client_error"),
        pytest.param(False, True, 1, False, True, id="describe_stack_no_credentials"),
        pytest.param(False, False, 0, False, True, id="unexpected_stacks"),
        pytest.param(False, False, 1, True, True, id="invalid_stack_status"),
    ],
)
def test_AWSCloudformationHandler_init(
    describe_stack_client_error,
    describe_stack_no_credentials,
    stack_num,
    stack_status_invalid,
    expected_error,
):
    mock_id = "mock_id"
    mock_region = "us-west-2"
    mock_client = Mock()
    mock_describe_stacks = Mock()
    mock_client.describe_stacks = mock_describe_stacks
    if describe_stack_client_error:
        mock_describe_stacks.side_effect = ClientError(
            {"Error": {"Code": "500", "Message": "Mock"}}, "Mock"
        )
    if describe_stack_no_credentials:
        mock_describe_stacks.side_effect = NoCredentialsError()
    stack_list = []
    for _ in range(stack_num):
        stack_list.append(
            {
                "StackName": "mock_stack",
                "StackStatus": "CREATE_COMPLETE"
                if not stack_status_invalid
                else "CREATE_FAILED",
            }
        )
    mock_describe_stacks.return_value = {"Stacks": stack_list}

    with patch.multiple(
        "anyscale.utils.cloud_update_utils", _client=Mock(return_value=mock_client)
    ):
        if expected_error:
            with pytest.raises(ClickException) as e:
                AWSCloudformationHandler(mock_id, mock_region, CloudSetupLogger())
            if describe_stack_client_error:
                e.match("Failed to describe cloudformation stack")
            elif describe_stack_no_credentials:
                e.match("Unable to locate AWS credentials.")
            elif stack_num != 1:
                e.match("unexpected number of cloudformation stacks")
            elif stack_status_invalid:
                e.match("stack is in a stable state")
        else:
            handler = AWSCloudformationHandler(mock_id, mock_region, CloudSetupLogger())
            assert handler.cfn_client == mock_client
            assert handler.region == mock_region
            assert handler.aws_cloudformation_stack_id == mock_id


@pytest.mark.parametrize("get_resource_failed", [True, False])
@mock_cloudformation
def test_AWSCloudformationHandler_get_resource(get_resource_failed):
    region = "us-west-2"
    mock_stack_name = "mock_cloudformation_stack"
    mock_logical_id = "mock_resource"
    cloudformation = boto3.client("cloudformation", region_name=region)
    cfn = cloudformation.create_stack(
        StackName=mock_stack_name,
        TemplateBody=json.dumps(
            {
                "Resources": {
                    "mock_resource": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "InstanceType": "t2.micro",
                            "ImageId": "ami-0b69ea66ff7391e80",
                        },
                    }
                }
            }
        ),
    )
    handler = AWSCloudformationHandler(cfn["StackId"], region, CloudSetupLogger())

    if get_resource_failed:
        with pytest.raises(ClickException) as e:
            handler.get_resource("no_such_id")
        e.match("Failed to get resource")
    else:
        resource = handler.get_resource(mock_logical_id)
        assert resource.get("LogicalResourceId") == mock_logical_id


@pytest.mark.parametrize(
    ("detection_failed", "timeout"),
    [
        pytest.param(True, False, id="detection_failed"),
        pytest.param(False, True, id="timeout"),
    ],
)
def test_AWSCloudformationHandler_detect_drift(detection_failed: bool, timeout: bool):
    # we don't use moto here since moto doesn't support drift detection
    mock_id = "mock_id"
    mock_describe_stack_drift_detection_status = Mock(
        return_value={"DetectionStatus": "DETECTION_COMPLETE",}
    )
    mock_detection_failed_reason = "mock"
    if detection_failed:
        mock_describe_stack_drift_detection_status = Mock(
            return_value={
                "DetectionStatus": "DETECTION_FAILED",
                "DetectionStatusReason": mock_detection_failed_reason,
            }
        )
    elif timeout:
        mock_describe_stack_drift_detection_status = Mock(
            return_value={"DetectionStatus": "DETECTION_IN_PROGRESS",}
        )
    mock_drifts = Mock()
    mock_cfn_client = MagicMock(
        detect_stack_drift=Mock(return_value={"StackDriftDetectionId": mock_id}),
        describe_stack_drift_detection_status=mock_describe_stack_drift_detection_status,
        describe_stack_resource_drifts=Mock(
            return_value={"StackResourceDrifts": mock_drifts}
        ),
        describe_stacks=Mock(
            return_value={
                "Stacks": [{"StackStatus": "CREATE_COMPLETE", "StackName": mock_id}]
            }
        ),
    )
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        _client=Mock(return_value=mock_cfn_client),
        DETECT_DRIFT_TIMEOUT_SECONDS=1,
    ):
        cfn_handler = AWSCloudformationHandler(mock_id, "us-west-2", CloudSetupLogger())
        if not detection_failed and not timeout:
            assert cfn_handler.detect_drift() == mock_drifts
        else:
            with pytest.raises(ClickException) as e:
                cfn_handler.detect_drift()
            if detection_failed:
                assert e.match(mock_detection_failed_reason)
            elif timeout:
                e.match("timeout")


@pytest.mark.parametrize(
    ("no_drift", "append_failed", "extract_failed"),
    [
        pytest.param(True, False, False, id="no_drift"),
        pytest.param(False, True, False, id="append_failed"),
        pytest.param(False, False, True, id="extract_failed"),
        pytest.param(False, False, False, id="happy_path"),
    ],
)
def test_AWSCloudformationHandler_resolve_drift(
    no_drift: bool, append_failed: bool, extract_failed: bool, capsys
):
    mock_extract_drifted_statements = Mock(
        return_value=[] if no_drift else [MagicMock()]
    )
    if extract_failed:
        mock_extract_drifted_statements.side_effect = ClickException("mock")
    mock_append_statements_to_customer_drifts_policy = Mock()
    if append_failed:
        mock_append_statements_to_customer_drifts_policy.side_effect = ClickException(
            "mock"
        )
    mock_client = MagicMock(
        describe_stacks=Mock(
            return_value={
                "Stacks": [{"StackStatus": "CREATE_COMPLETE", "StackName": "mock_id"}]
            }
        ),
    )

    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        extract_drifted_statements=mock_extract_drifted_statements,
        append_statements_to_customer_drifts_policy=mock_append_statements_to_customer_drifts_policy,
        _client=Mock(return_value=mock_client),
    ):

        cfn_handler = AWSCloudformationHandler("mock", "us-west-2", CloudSetupLogger())
        result = cfn_handler.resolve_drift(MagicMock())
    _, stdout = capsys.readouterr()
    if no_drift:
        assert "No drifted statements found." in stdout
        assert result is True
    elif append_failed or extract_failed:
        assert "Failed to resolve drift" in stdout
        assert result is False
    else:
        assert "Drifted statements have been appended to the policy" in stdout
        assert result is True


@mock_cloudformation
def test_AWSCloudformationHandler_get_template_body():
    region = "us-west-2"
    mock_stack_name = "mock_cloudformation_stack"

    mock_template_body = json.dumps(
        {
            "Resources": {
                "mock_resource": {
                    "Type": "AWS::EC2::Instance",
                    "Properties": {
                        "InstanceType": "t2.micro",
                        "ImageId": "ami-0b69ea66ff7391e80",
                    },
                }
            }
        }
    )

    cloudformation = boto3.client("cloudformation", region_name=region)
    cfn = cloudformation.create_stack(
        StackName=mock_stack_name, TemplateBody=mock_template_body,
    )

    cfn_handler = AWSCloudformationHandler(cfn["StackId"], region, CloudSetupLogger())
    template_body = cfn_handler.get_template_body()
    assert json.dumps(template_body) == mock_template_body


@pytest.mark.parametrize(
    (
        "create_change_set_error",
        "create_change_set_timeout",
        "update_stack_error",
        "update_stack_timeout",
    ),
    [
        pytest.param(True, False, False, False, id="create_change_set_error"),
        pytest.param(False, True, False, False, id="create_change_set_timeout"),
        pytest.param(False, False, True, False, id="update_stack_error"),
        pytest.param(False, False, False, True, id="update_stack_timeout"),
        pytest.param(False, False, False, False, id="happy-path"),
    ],
)
def test_AWSCloudformationHandler_update_cloudformation_stack(
    create_change_set_error,
    create_change_set_timeout,
    update_stack_error,
    update_stack_timeout,
):
    mock_stack_name = "mock_stack_name"
    mock_region = "mock_region"
    mock_change_set_id = "mock_change_set_id"
    mock_template_body, mock_parameters = _get_cloudformation_template_and_parameters()
    mock_change_set = {
        "Status": "CREATE_COMPLETE",
        "StackId": "mock_stack_id",
    }
    if create_change_set_error:
        mock_change_set = {
            "Status": "FAILED",
            "StatusReason": "mock_status_reason",
        }
    elif create_change_set_timeout:
        mock_change_set = {
            "Status": "CREATE_IN_PROGRESS",
        }
    mock_stacks = {
        "Stacks": [{"StackName": mock_stack_name, "StackStatus": "UPDATE_COMPLETE",}]
    }
    mock_describe_stacks = Mock(return_value=mock_stacks)
    if update_stack_error:
        mock_stacks["Stacks"][0]["StackStatus"] = "UPDATE_ROLLBACK_COMPLETE"
    elif update_stack_timeout:
        mock_stacks["Stacks"][0]["StackStatus"] = "CREATE_COMPLETE"

    mock_cfn_client = MagicMock(
        get_template=Mock(return_value={"TemplateBody": mock_template_body}),
        create_change_set=Mock(return_value={"Id": mock_change_set_id}),
        describe_change_set=Mock(return_value=mock_change_set),
        delete_change_set=Mock(),
        execute_change_set=Mock(),
        describe_stacks=mock_describe_stacks,
    )
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        _client=Mock(return_value=mock_cfn_client),
        CREATE_CHANGE_SET_TIMEOUT_SECONDS=1,
        UPDATE_CLOUDFORMATION_STACK_TIMEOUT_SECONDS=1,
    ):
        cfn_handler = AWSCloudformationHandler(
            mock_stack_name, mock_region, CloudSetupLogger()
        )
        if (
            create_change_set_error
            or create_change_set_timeout
            or update_stack_error
            or update_stack_timeout
        ):
            with pytest.raises(ClickException) as e:
                cfn_handler.update_cloudformation_stack(
                    mock_template_body, mock_parameters, True, timeout_seconds=1
                )
            if create_change_set_error:
                assert e.match("Failed to create change set")
                mock_cfn_client.delete_change_set.assert_called_once_with(
                    ChangeSetName=mock_change_set_id
                )
                mock_cfn_client.execute_change_set.assert_not_called()
            elif create_change_set_timeout:
                assert e.match("Timeout when creating change set")
                mock_cfn_client.execute_change_set.assert_not_called()
            elif update_stack_error:
                assert e.match("Failed to execute change set")
                mock_cfn_client.execute_change_set.assert_called_once_with(
                    ChangeSetName=mock_change_set_id
                )
            elif update_stack_timeout:
                assert e.match("Timeout when executing change set")
                mock_cfn_client.execute_change_set.assert_called_once_with(
                    ChangeSetName=mock_change_set_id
                )
        else:
            cfn_handler.update_cloudformation_stack(
                mock_template_body, mock_parameters, True,
            )


@pytest.mark.parametrize(
    ("stack_version", "cli_version", "expected_result"),
    [
        pytest.param("0.0.2", "0.0.2", True, id="version-match"),
        pytest.param("0.0.1", "0.0.2", False, id="version-mismatch"),
        pytest.param(None, "0.0.2", False, id="no-stack-version"),
        pytest.param(None, "0.0.0-dev", False, id="dev-cli-1"),
        pytest.param("0.0.2", "0.0.0-dev", False, id="dev-cli-2"),
    ],
)
def test_validate_stack_version(stack_version, cli_version, expected_result):
    mock_parameters = []
    if stack_version:
        mock_parameters.append(
            {"ParameterKey": "AnyscaleCLIVersion", "ParameterValue": stack_version}
        )
    with patch.multiple(
        "anyscale.utils.cloud_update_utils", anyscale_version=cli_version,
    ):
        validate_stack_version(mock_parameters)


@pytest.mark.parametrize("missing_parameter", [True, False])
def test_generate_updated_parameters_and_template(missing_parameter):
    mock_parameters = [
        {"ParameterKey": "AnyscaleCLIVersion", "ParameterValue": "0.0.1"},
    ]
    parameters_to_update = [
        {"ParameterKey": "AnyscaleCLIVersion", "ParameterValue": "0.0.2"},
    ]
    mock_missing_parameter_key = "MemoryDBRedisPort"
    if missing_parameter:
        parameters_to_update.append(
            {"ParameterKey": mock_missing_parameter_key, "ParameterValue": "6379"}
        )
    mock_template_body = "mock"
    mock_add_missing_parameters_to_template_body = Mock()
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        merge_parameters=Mock(),
        add_missing_parameters_to_template_body=mock_add_missing_parameters_to_template_body,
    ):
        generate_updated_parameters_and_template(
            mock_parameters, mock_template_body, parameters_to_update
        )
        if missing_parameter:
            mock_add_missing_parameters_to_template_body.assert_called_once_with(
                mock_template_body, {mock_missing_parameter_key}
            )


def test_update_template_with_memorydb():
    mock_template_body = """
Resources:
Outputs:
"""
    assert (
        update_template_with_memorydb(mock_template_body, 2)
        == """
Resources:
  MemoryDBSubnetGroup:
    Type: AWS::MemoryDB::SubnetGroup
    Properties:
      Description: Anyscale managed MemoryDB subnet group
      SubnetGroupName: !Ref AWS::StackName
      SubnetIds:
        - !Ref Subnet0
        - !Ref Subnet1
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID

  MemoryDBParameterGroup:
    Type: AWS::MemoryDB::ParameterGroup
    Properties:
      Description: Parameter group for anyscale managed MemoryDB
      Family: memorydb_redis7
      ParameterGroupName:  !Ref AWS::StackName
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID

  MemoryDB:
    Type: AWS::MemoryDB::Cluster
    Properties:
      ACLName: open-access
      Description: Anyscale managed MemoryDB
      ClusterName: !Ref AWS::StackName
      NodeType: db.t4g.small
      Port: !Ref MemoryDBRedisPort
      SubnetGroupName: !Ref MemoryDBSubnetGroup
      SecurityGroupIds:
        - !Ref AnyscaleSecurityGroup
      EngineVersion: "7.0"
      ParameterGroupName: !Ref MemoryDBParameterGroup
      TLSEnabled: true
      Tags:
        - Key: anyscale-cloud-id
          Value: !Ref CloudID

Outputs:
  MemoryDB:
    Description: MemoryDB cluster
    Value:
      Fn::ToJsonString:
        arn: !GetAtt MemoryDB.ARN
        ClusterEndpointAddress: !GetAtt MemoryDB.ClusterEndpoint.Address

"""
    )


@pytest.mark.parametrize(
    (
        "stack_version_match",
        "parameter_up_to_date",
        "has_role_drift",
        "resolve_drift_failed",
    ),
    [
        pytest.param(False, False, False, False, id="happy-path"),
        pytest.param(True, False, False, False, id="version-match"),
        pytest.param(False, True, False, False, id="parameters-up-to-date"),
        pytest.param(False, False, True, False, id="has-role-drift"),
        pytest.param(False, False, True, True, id="resolve-drift-failed"),
    ],
)
def test_update_iam_role(
    stack_version_match: bool,
    parameter_up_to_date: bool,
    has_role_drift: bool,
    resolve_drift_failed: bool,
    capsys,
):
    region = "us-west-2"
    mock_cfn_id = "mock_id"
    mock_handler = Mock()
    mock_handler.get_cloudformation_stack = MagicMock(Parameters=Mock())
    mock_handler.detect_drift = Mock()
    mock_handler.resolve_drift = Mock(return_value=not resolve_drift_failed)
    mock_handler.update_cloudformation_stack = Mock()
    mock_handler.get_template_body = Mock(return_value="mock_template")

    mock_validate_stack_version = Mock(return_value=not stack_version_match)
    mock_is_template_policy_documents_up_to_date = Mock(
        return_value=parameter_up_to_date
    )
    mock_generate_updated_parameters_and_template = Mock(return_value=(Mock(), Mock()))
    mock_confirm = Mock()
    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        AWSCloudformationHandler=Mock(return_value=mock_handler),
        validate_stack_version=mock_validate_stack_version,
        is_template_policy_documents_up_to_date=mock_is_template_policy_documents_up_to_date,
        extract_cross_account_iam_role_drift=Mock(
            return_value=MagicMock(LogicalResourceId="mock") if has_role_drift else None
        ),
        generate_updated_parameters_and_template=mock_generate_updated_parameters_and_template,
        confirm=mock_confirm,
    ):
        update_iam_role(region, mock_cfn_id, CloudSetupLogger(), True)
        mock_validate_stack_version.assert_called_once()
        _, stdout = capsys.readouterr()

        if stack_version_match:
            mock_is_template_policy_documents_up_to_date.assert_not_called()
            assert "Cloud is already up-to-date." in stdout
        elif parameter_up_to_date:
            mock_is_template_policy_documents_up_to_date.assert_called_once()
            mock_handler.detect_drift.assert_not_called()
            assert "No inline policy changes" in stdout
        else:
            mock_is_template_policy_documents_up_to_date.assert_called_once()
            mock_handler.detect_drift.assert_called_once()
            mock_handler.get_template_body.assert_called_once()
            mock_generate_updated_parameters_and_template.assert_called_once()
            mock_handler.update_cloudformation_stack.assert_called_once()

            if has_role_drift:
                assert "Drfits detected" in stdout
                mock_handler.resolve_drift.assert_called_once()
                if resolve_drift_failed:
                    assert mock_confirm.call_count == 2
                else:
                    mock_confirm.assert_called_once()
            else:
                mock_handler.resolve_drift.assert_not_called()
                mock_confirm.assert_not_called()


@pytest.mark.parametrize("memorydb_exists", [True, False])
def test_get_or_create_memorydb(memorydb_exists: bool):
    mock_handler = Mock()
    mock_stack = MagicMock(Parameters=[])
    mock_template = "mock_template"
    if memorydb_exists:
        mock_template += " AWS::MemoryDB::Cluster"

    mock_handler.get_cloudformation_stack = Mock(return_value=mock_stack)
    mock_handler.get_template_body = Mock(return_value=mock_template)
    mock_handler.update_cloudformation_stack = Mock()
    mock_memorydb_id = "mock_id"
    mock_handler.get_resource = Mock(
        return_value={"PhysicalResourceId": mock_memorydb_id}
    )

    mock_generate_updated_parameters_and_template = Mock(return_value=(Mock(), Mock()))
    mock_update_template_with_memorydb = Mock()

    with patch.multiple(
        "anyscale.utils.cloud_update_utils",
        AWSCloudformationHandler=Mock(return_value=mock_handler),
        generate_updated_parameters_and_template=mock_generate_updated_parameters_and_template,
        update_template_with_memorydb=mock_update_template_with_memorydb,
    ):
        result = get_or_create_memorydb(
            "mock_stack_name", "mock_region", 2, CloudSetupLogger(), True
        )
        assert result == mock_memorydb_id
        mock_handler.get_resource.assert_called_once()
        if memorydb_exists:
            mock_handler.update_cloudformation_stack.assert_not_called()
            mock_generate_updated_parameters_and_template.assert_not_called()
            mock_update_template_with_memorydb.assert_not_called()
        else:
            mock_handler.update_cloudformation_stack.assert_called_once()
            mock_generate_updated_parameters_and_template.assert_called_once()
            mock_update_template_with_memorydb.assert_called_once()



from typing import Optional
from unittest.mock import ANY, Mock

import pytest

from anyscale.cluster_compute import (
    get_cluster_compute_from_name,
    get_default_cluster_compute,
    get_selected_cloud_id_or_default,
    register_compute_template,
)
from anyscale.sdk.anyscale_client import ArchiveStatus, CreateComputeTemplate
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)


def test_get_default_cluster_compute():
    mock_api_client = Mock()
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.get_default_compute_config = Mock(
        return_value=Mock(result="mock_config_obj")
    )
    mock_default_config_obj = {
        "cloud_id": "cld_1",
        "max_workers": 10,
        "region": "us-west-2",
        "allowed_azs": ["us-west-2a", "us-west-2b"],
        "head_node_type": {
            "name": "head-1",
            "instance_type": "m5.2xlarge",
            "aws_advanced_configurations": None,
            "gcp_advanced_configurations": None,
            "resources": None,
        },
        "worker_node_types": [],
    }
    mock_anyscale_api_client.get_default_compute_config = Mock(
        return_value=Mock(result=mock_default_config_obj)
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result="mock_compute_template")
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=Mock(result=Mock(id="test_cloud_id"))
    )

    assert (
        get_default_cluster_compute(
            "mock_cloud_name", None, mock_api_client, mock_anyscale_api_client
        )
        == "mock_compute_template"
    )
    mock_anyscale_api_client.get_default_compute_config.assert_called_once_with(
        "test_cloud_id"
    )


def test_register_compute_template():
    mock_api_client = Mock()
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result=Mock(id="mock_template_id"))
    )

    assert (
        register_compute_template("mock_template_obj", mock_api_client).id
        == "mock_template_id"
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post.assert_called_once_with(
        create_compute_template=CreateComputeTemplate(
            name=ANY, config="mock_template_obj", anonymous=True,
        )
    )


@pytest.mark.parametrize("version", [None, 2])
def test_get_cluster_compute_from_name(version: Optional[int]):
    mock_api_client = Mock()
    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(results=[Mock(id="mock_id")])
    )
    assert (
        get_cluster_compute_from_name("mock_name", mock_api_client, version).id
        == "mock_id"
    )

    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_once_with(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": "mock_name"},
            include_anonymous=True,
            archive_status=ArchiveStatus.ALL,
            version=version,
        )
    )


def test_get_selected_cloud_id_or_default():
    mock_api_client = Mock()
    mock_anyscale_api_client = Mock()

    mock_cloud = Mock(id="mock_cloud_id")
    mock_cloud.name = "mock_cloud_name"

    # Test getting cloud from cluster compute id
    assert (
        get_selected_cloud_id_or_default(
            api_client=mock_api_client,
            anyscale_api_client=mock_anyscale_api_client,
            cluster_compute_id="mock_cluster_compute_id",
        )
        == mock_anyscale_api_client.get_cluster_compute(
            "mock_cluster_compute_id"
        ).result.config.cloud_id
    )

    # Test getting cloud from cluster compute config
    cluster_compute_dict = {
        "cloud_id": "mock_cloud_id",
        "max_workers": 10,
        "region": "us-west-2",
        "allowed_azs": ["us-west-2a", "us-west-2b"],
        "head_node_type": {
            "name": "head-1",
            "instance_type": "m5.2xlarge",
            "aws_advanced_configurations": None,
            "gcp_advanced_configurations": None,
            "resources": None,
        },
        "worker_node_types": [],
    }
    cluster_compute_config = ClusterComputeConfig(**cluster_compute_dict)
    assert (
        get_selected_cloud_id_or_default(
            api_client=mock_api_client,
            anyscale_api_client=mock_anyscale_api_client,
            cluster_compute_config=cluster_compute_config,
        )
        == "mock_cloud_id"
    )

    # Test getting cloud from cloud_id
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
        return_value=Mock(result=mock_cloud)
    )
    assert (
        get_selected_cloud_id_or_default(
            api_client=mock_api_client,
            anyscale_api_client=mock_anyscale_api_client,
            cloud_id="mock_cloud_id",
        )
        == "mock_cloud_id"
    )
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
        cloud_id="mock_cloud_id"
    )

    # Test getting cloud from cloud_name
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=Mock(result=mock_cloud)
    )
    assert (
        get_selected_cloud_id_or_default(
            api_client=mock_api_client,
            anyscale_api_client=mock_anyscale_api_client,
            cloud_name="mock_cloud_name",
        )
        == "mock_cloud_id"
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once_with(
        cloud_name_options={"name": "mock_cloud_name"}
    )

    # Test getting cloud from default cluster compute
    mock_anyscale_api_client.get_default_cluster_compute = Mock(
        return_value=Mock(result=Mock(config=Mock(cloud_id="mock_default_cloud_id")))
    )
    assert (
        get_selected_cloud_id_or_default(
            api_client=mock_api_client, anyscale_api_client=mock_anyscale_api_client
        )
        == "mock_default_cloud_id"
    )
    mock_anyscale_api_client.get_default_cluster_compute.assert_called_once_with()



import asyncio
from http import HTTPStatus
from unittest.mock import Mock, patch

from aiohttp import ClientConnectionError
from asynctest import CoroutineMock
import pytest

from anyscale.api_utils import job_logs_util
from anyscale.api_utils.exceptions.log_retrieval_errors import (
    UnsupportedLogRetrievalMethodError,
)
from anyscale.api_utils.job_logs_util import (
    _get_job_logs_from_storage_bucket,
    _get_logs_from_active_job_run,
    _get_logs_from_running_production_job,
)
from anyscale.api_utils.job_util import _get_job_run_id
from anyscale.api_utils.logs_util import (
    _download_log_from_ray_json_response,
    _download_logs_concurrently,
)
from anyscale.controllers.logs_controller import DEFAULT_PARALLELISM
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from tests.api_utils.conftest import get_mock_api_exception, get_mock_base_api


@pytest.mark.asyncio()
async def test_get_logs_from_active_job_run():
    mock_stream_url = "https://valid.ray.dashboard/logs"
    mock_log_stream = Mock(http_url=mock_stream_url)
    mock_base_api = get_mock_base_api(
        {BaseApi.get_job_logs_stream.__name__: Mock(result=mock_log_stream)}
    )
    mock_logs = "logs"
    job_run_id = "job_123"

    with patch.object(
        job_logs_util,
        _download_log_from_ray_json_response.__name__,
        CoroutineMock(return_value=mock_logs),
    ) as mock_download_log_from_ray_json_response:
        assert (
            await _get_logs_from_active_job_run(mock_base_api, job_run_id) == mock_logs
        )
    mock_base_api.get_job_logs_stream.assert_called_once_with(job_id=job_run_id)
    mock_download_log_from_ray_json_response.assert_awaited_once_with(mock_stream_url)


# Indirectly tests `wrap_as_unsupported_log_retrieval_method_error()`
@pytest.mark.asyncio()
async def test_get_logs_from_active_job_run_throws_exception_anyscale_api():
    # Test expected ApiException
    error_message = "Cluster is not active."
    mock_api_exception = get_mock_api_exception(HTTPStatus.BAD_REQUEST, error_message)
    mock_base_api = get_mock_base_api(
        {BaseApi.get_job_logs_stream.__name__: mock_api_exception}
    )
    job_run_id = "job_123"
    with pytest.raises(UnsupportedLogRetrievalMethodError, match=error_message):
        await _get_logs_from_active_job_run(mock_base_api, job_run_id)

    # Test unexpected ApiException
    mock_api_exception.status = HTTPStatus.NOT_FOUND
    mock_api_exception.body = "Not found"
    with pytest.raises(Exception, match="Not found"):
        await _get_logs_from_active_job_run(mock_base_api, job_run_id)


# Indirectly tests `wrap_job_run_log_not_retrievable_on_active_cluster_error()`
@pytest.mark.asyncio()
async def test_get_logs_from_active_job_run_throws_exception_ray_dashboard_api():
    mock_base_api = get_mock_base_api(
        {
            BaseApi.get_job_logs_stream.__name__: Mock(
                result=Mock(http_url="https://valid.ray.dashboard/logs")
            )
        }
    )
    job_run_id = "job_123"

    for exception_type in [ClientConnectionError, asyncio.TimeoutError]:
        with patch.object(
            job_logs_util,
            _download_log_from_ray_json_response.__name__,
            CoroutineMock(side_effect=exception_type),
        ), pytest.raises(
            exception_type,
            match="only supported if the Anyscale client \\(SDK / CLI\\) can reach the cluster",
        ):
            await _get_logs_from_active_job_run(mock_base_api, job_run_id)


@pytest.mark.asyncio()
async def test_get_job_logs_from_storage_bucket():
    job_run_id = "job_123"
    mock_base_api = get_mock_base_api(
        {
            BaseApi.get_job_logs_download.__name__: Mock(
                result=Mock(log_chunks=[Mock(chunk_url="url1"), Mock(chunk_url="url2")])
            )
        }
    )
    mock_get_job_run_id = Mock(return_value=job_run_id)
    mock_logs = "a\nb"
    mock_download_logs_concurrently = CoroutineMock(return_value=mock_logs)
    to_patch = {
        _get_job_run_id.__name__: mock_get_job_run_id,
        _download_logs_concurrently.__name__: mock_download_logs_concurrently,
    }
    with patch.multiple(job_logs_util, **to_patch):
        assert (
            await _get_job_logs_from_storage_bucket(
                mock_base_api, job_run_id=job_run_id
            )
            == "a\nb"
        )
    mock_get_job_run_id.assert_called_once_with(
        mock_base_api, job_id=None, job_run_id=job_run_id
    )
    mock_base_api.get_job_logs_download.assert_called_once_with(
        job_id=job_run_id, all_logs=True
    )
    mock_download_logs_concurrently.assert_awaited_once_with(
        ["url1", "url2"], DEFAULT_PARALLELISM
    )

    # Test error from Anyscale API
    mock_download_logs_concurrently.reset_mock()
    error_message = "Cannot download"
    mock_base_api.get_job_logs_download.side_effect = get_mock_api_exception(
        HTTPStatus.BAD_REQUEST, error_message
    )
    with patch.multiple(job_logs_util, **to_patch), pytest.raises(
        UnsupportedLogRetrievalMethodError, match=error_message
    ):
        await _get_job_logs_from_storage_bucket(mock_base_api, job_run_id=job_run_id)
    mock_download_logs_concurrently.assert_not_awaited()


@pytest.mark.asyncio()
async def test_get_logs_from_running_production_job():
    job_id = "prodjob_123"
    job_run_id = "job_123"
    mock_base_api = Mock()
    mock_get_job_run_id = Mock(return_value=job_run_id)
    mock_logs = "a\nb"
    mock_get_logs_from_active_job_run = CoroutineMock(return_value=mock_logs)
    to_patch = {
        _get_job_run_id.__name__: mock_get_job_run_id,
        _get_logs_from_active_job_run.__name__: mock_get_logs_from_active_job_run,
    }
    with patch.multiple(job_logs_util, **to_patch):
        assert (
            await _get_logs_from_running_production_job(mock_base_api, job_id)
            == mock_logs
        )
    mock_get_job_run_id.assert_called_once_with(mock_base_api, job_id=job_id)
    mock_get_logs_from_active_job_run.assert_awaited_once_with(
        mock_base_api, job_run_id, True
    )
